import bpy
from .BlenderWidgetManager import define_widget, define_widget_button

class BlenderDummyWidget:
  def __init__(self, parent, drawfunc):
    self.parent = parent  
    self.drawfunc = drawfunc
  def __getattr__(self, attr):
    raise NotImplementedError(attr)
  def draw(self, context, layout):
    self.drawfunc(context, layout)

class BlenderWrapWidget:
  def __init__(self, parent):
    self.parent = parent  
  def __getattr__(self, attr):
    raise NotImplementedError(attr)
    
class BlenderEmptyWidget:
  def __init__(self, parent):
    self.parent = parent  
  def show(self): pass
  def hide(self): pass
  def __getattr__(self, attr):
    raise NotImplementedError(attr)

class BlenderWidget:
  def draw(self, context, layout):    
    raise NotImplementedError
      
class BlenderLayoutWidget(BlenderWidget):    
  def __init__(self, name=None, buttons = []):
    self.name = name
    self.children = []
    self.pre_buttons = [w for w in buttons if w.layout == "before"]
    self.post_buttons = [w for w in buttons if w.layout == "after"]
  def __getattr__(self, attr):
    raise NotImplementedError(attr)
  def draw(self, context, layout):    
    if self.name is not None:
      layout = layout.row()
      layout.label(self.name)
    layout=layout.box()
    if not self.children and not self.pre_buttons and not self.post_buttons: return
    """
    #TODO: column/row layout (__init__ option, determined from the spyderform) 
    """
    for but in self.pre_buttons:
      but.draw(context, layout)  
    for child in self.children:
      child.draw(context, layout)
    for but in self.post_buttons:
      but.draw(context, layout)        
  def show(self):
    for child in self.children:
      if hasattr(child, "show"): child.show()
  def hide(self):
    for child in self.children:
      if hasattr(child, "hide"): child.hide()
        
class BlenderPlaceholderWidget(BlenderWidget):    
  def __init__(self, name, typ):
    self.name = name
    self.typ = typ
    self.value = None
  def listen(self, callback): pass
  def set(self, value): self.value = value  
  def get(self): return self.value
  def draw(self, context, layout):    
    txt = "Placeholder: name '%s', type '%s', value '%s'" % (self.name, self.typ, self.value)
    layout.label(txt)
 
class BlenderLabelWidget(BlenderWidget):    
  def __init__(self, text):
    self.text = text    
  def draw(self, context, layout):        
    layout.label(self.text)

class BlenderButtonWidget(BlenderWidget):    
  def __init__(self, txt, layout=None):
    self.txt = txt
    self.layout = layout # "before" or "after"
    self.widget_id = define_widget_button(self.press)
    self._listeners = []
  def listen(self, callback): 
    self._listeners.append(callback)
  def unlisten(self, callback): 
    self._listeners.remove(callback)
  def press(self):  
    for callback in self._listeners: 
      callback()
  def draw(self, context, layout):        
    layout.operator(self.widget_id, self.txt)

class BlenderIntWidget(BlenderWidget):    
  def __init__(self, name):
    self.name = name
    self.value = None
    self.widget_id = define_widget(bpy.props.IntProperty, "intprop", self.get2, self.set)
    self._listeners = []
  def listen(self, callback): 
    self._listeners.append(callback)
  def unlisten(self, callback): 
    self._listeners.remove(callback)
  def block(self):
    self._blockedcount += 1
  def unblock(self):
    self._blockedcount -= 1
    if self._blockedcount < 0: self._blockedcount = 0    
  def set(self, value): 
    self.value = value  
    for callback in self._listeners: callback(self.value)
  def get(self): return self.value
  def get2(self): 
    if self.value is None: return 0
    return self.value
  def draw(self, context, layout):    
    layout.prop(context.scene, self.widget_id, text=self.name)

class BlenderFloatWidget(BlenderWidget):    
  def __init__(self, name):
    self.name = name
    self.value = None
    self.widget_id = define_widget(bpy.props.FloatProperty, "floatprop", self.get2, self.set)
    self._listeners = []
  def listen(self, callback): 
    self._listeners.append(callback)
  def unlisten(self, callback): 
    self._listeners.remove(callback)
  def block(self):
    self._blockedcount += 1
  def unblock(self):
    self._blockedcount -= 1
    if self._blockedcount < 0: self._blockedcount = 0    
  def set(self, value): 
    self.value = value  
    for callback in self._listeners: callback(self.value)
  def get(self): return self.value
  def get2(self): 
    if self.value is None: return 0.0
    return self.value
  def draw(self, context, layout):    
    layout.prop(context.scene, self.widget_id, text=self.name)
    
class BlenderStringWidget(BlenderWidget):    
  def __init__(self, name):
    self.name = name
    self.value = None
    self.widget_id = define_widget(bpy.props.StringProperty, "stringprop", self.get2, self.set)
    self._listeners = []
  def listen(self, callback): 
    self._listeners.append(callback)
  def unlisten(self, callback): 
    self._listeners.remove(callback)
  def block(self):
    self._blockedcount += 1
  def unblock(self):
    self._blockedcount -= 1
    if self._blockedcount < 0: self._blockedcount = 0    
  def set(self, value):     
    self.value = value  
    for callback in self._listeners: callback(self.value)
  def get(self): return self.value
  def get2(self): 
    if self.value is None: return ""
    return str(self.value)
  def draw(self, context, layout):    
    layout.prop(context.scene, self.widget_id, text=self.name)
    
class BlenderBoolWidget(BlenderWidget):    
  def __init__(self, name):
    self.name = name
    self.value = None
    self.widget_id = define_widget(bpy.props.BoolProperty, "boolprop", self.get2, self.set)
    self._listeners = []
  def listen(self, callback): 
    self._listeners.append(callback)
  def unlisten(self, callback): 
    self._listeners.remove(callback)
  def block(self):
    self._blockedcount += 1
  def unblock(self):
    self._blockedcount -= 1
    if self._blockedcount < 0: self._blockedcount = 0    
  def set(self, value): 
    self.value = value  
    for callback in self._listeners: callback(self.value)
  def get(self): return self.value
  def get2(self): 
    if self.value is None: return False
    return self.value
  def draw(self, context, layout):    
    layout.prop(context.scene, self.widget_id, text=self.name)

class BlenderOptionWidget(BlenderWidget):    
  """
  Defines an multiple-choice string widget
  """
  def __init__(self, name, options, option_names = None, option_descriptions = None):
    self.name = name
    self.value = None #string!
    if option_names is not None: assert len(option_names) == len(options)
    if option_descriptions is not None: assert len(option_descriptions) == len(options)
    self.options = options    
    self.option_names = option_names
    self.option_descriptions = option_descriptions
    def get_items2(dummy, dummy2):
      return self.get_items()
    self.widget_id = define_widget(bpy.props.EnumProperty, 
     "enumprop", self.get_index, self.set_index, items = get_items2, name=self.name)
    self._listeners = []
  def listen(self, callback): 
    self._listeners.append(callback)
  def unlisten(self, callback): 
    self._listeners.remove(callback)
  def block(self):
    self._blockedcount += 1
  def unblock(self):
    self._blockedcount -= 1
    if self._blockedcount < 0: self._blockedcount = 0    
  def set(self, value): 
    assert value is None or value in self.options, (value, self.options)
    self.value = value  
    for callback in self._listeners: callback(self.value)
  def get(self): 
    return self.value
  def get_index(self): 
    if self.value is None: return 0
    return self.options.index(self.value) + 1
  def set_index(self, index):   
    if index == 0: 
      self.value = None
    else:
      self.value = self.options[index-1]
    for callback in self._listeners: callback(self.value)  
  def get_items(self):
    items = [("<empty>","","Value is not defined",0)]
    op = self.options
    opn = self.option_names
    if opn is None: opn = op
    opd = self.option_descriptions
    if opd is None: opd = opn
    opnr = range(1,len(op)+1)
    items2 = list(zip(op,opn,opd,opnr))
    return items + items2    
    
  def draw(self, context, layout):    
    layout.prop(context.scene, self.widget_id, text=self.name)
    