"""
Converts a form into a comparator object
A comparator object supports indexing and/or attribute access at the lowest level
However, all initial values are set to None

You can change only initial values, and only at the lowest level 
 (forms that have no subforms)
Files are recognized as a special case: their values are set with a File object,
  but actually the filename is stored


"""

from functools import partial

class comparator_element(object):
  def __init__(self, is_file):
    self._value = None
    self._listeners = []
    self._is_file = is_file
  def _get(self):
    return self._value
  def _listen(self, listener):
    self._listeners.append(listener)
  def _unlisten(self, listener):
    if listener in self._listeners: 
      self._listeners.remove(listener)    
      return True
    else: return False
  def _set(self, value):
    if value != self._value:
      self._do_set(value)
      for listener in self._listeners:
        listener(self._value)
    else:
      self._do_set(value)
  def _do_set(self, value):    
    if self._is_file and value is not None:
      self._value = value.name
    else:
      self._value = value
    
  
class comparator(object):
  def __init__(self, form, has_parent = False):
    self._form = None
    self._value = None
    self._length = 0
    self._struct_properties = None
    self._element_properties = None
    if form is None: 
      assert has_parent == True
      return
    import Spyder, spyder
    assert isinstance(form, spyder.core.spyderform)
    if hasattr(form, "type") and form.type == "none":
      assert has_parent == True
      return
    self._form = form
    membernames = form.get_membernames()
    if membernames is None and self._form.arraycount == 0: 
      assert has_parent == True
      assert form.typename is not None
      is_file = False 
      if hasattr(form, "type") and form.type == "file": 
        is_file = True
      if hasattr(form, "spydertype") and form.spydertype is not None:
        spydertype = form.spydertype
      else:
        spydertype = getattr(Spyder, form.typename)
      try: 
        if issubclass(spydertype, Spyder.File):
          is_file = True
      except TypeError:
        pass        
      self._is_file = is_file
      return
    if self._form.arraycount > 0:
      assert hasattr(form, "length") and form.length is not None
      membernames = [str(v) for v in range(form.length)]
      getfunc = lambda v: form.__getitem__(int(v))
    else:
      getfunc = partial(getattr, form)
    self._element_properties = {}
    self._struct_properties = {}
    for membername in membernames:
      if membername is None: continue
      f = comparator(getfunc(membername), True)
      if f._element_properties is not None: 
        self._struct_properties[membername] = f
      elif f._form is not None:         
        ff = comparator_element(f._is_file)
        self._element_properties[membername] = ff
      
  def __getitem__(self, key):    
    if not isinstance(key, int): raise TypeError
    if not key >= 0: raise IndexError
    if not self._form.arraycount > 0: raise TypeError
    try:
      ret = self._struct_properties[str(key)]
    except KeyError:    
      try:
        ret = self._element_properties[str(key)]
      except KeyError:
        raise IndexError
    if key > self._length: self._length = key
    return ret
  def __setitem__(self, key, value):    
    if not isinstance(key, int): raise TypeError
    if not key >= 0: raise IndexError
    if not self._form.arraycount > 0: raise TypeError
    kkey = str(key)
    assert kkey in self._element_properties
    self._element_properties[kkey]._set(value)

  def __getattr__(self, attr):
    try:
      return self._struct_properties[attr]
    except KeyError:
      try:
        return self._element_properties[attr]
      except KeyError:
        if self._value is None: 
          raise AttributeError(attr) 
        else:
          return getattr(self._value, attr)
  def __setattr__(self, attr, value):
    if attr.startswith("_"): 
      return object.__setattr__(self, attr, value)
    try:
      ret = self._element_properties[attr]._set(value)
      return ret
    except KeyError:
      if self._value is None: 
        raise AttributeError(attr) 
      else:
        return setattr(self._value, attr, value)

  def _set(self, value):
    if value is None: return
    if self._form.arraycount > 0:    
      for k in self._struct_properties:
        try:
          subvalue = value[int(k)]
        except IndexError:
          continue
        self._struct_properties[k]._set(subvalue)
      for k in self._element_properties:
        try:
          subvalue = value[int(k)]
        except IndexError:
          continue
        self._element_properties[k]._set(subvalue)

    else:
      if self._form._typetree.typename == None:
        getfunc = value.__getitem__
        exceptionclass = KeyError
      else:
        getfunc = partial(getattr, value)
        exceptionclass = AttributeError
      for k in self._struct_properties:
        try:
          subvalue = getfunc(k)
        except exceptionclass:
          continue
        self._struct_properties[k]._set(subvalue)          
      for k in self._element_properties:
        try:
          subvalue = getfunc(k)
        except exceptionclass:
          continue
        self._element_properties[k]._set(subvalue)

