import Spyder, spyder
import sys
from functools import partial
import copy as copy_module

"""
Models have a .get method with an optional "minimal" argument (default: False)
 If "minimal" = True, all non-default attributes are set to None

There are three concepts of "default":
- Top-level default (called "default")
  The default value of an object when its top level parent is constructed
- Direct parent default (called "pdefault")
  The default value of an object when its direct level parent is constructed
- Self default (called "sdefault")
  The default value of an object when the object is constructed without any arguments
  In Spyder, this property is never specified directly, it is just the composite dict of 
    the pdefaults of all its members (and always empty for arrays)

Example: 
Type TestType {
  ax = AxisSystem(x=(1,2,3))
}

TestType.ax (AxisSystem):
  top-level default: 
   origin=(0,0,0), x = (1,2,3), y=(0,1,0), z = (0,0,1)
  direct parent default: same as top-level
  self default:
   origin=(0,0,0), x = (1,0,0), y=(0,1,0), z = (0,0,1)

AxisSystem.x (Coordinate):
  top-level default: (1,0,0)
  direct parent default: same as top-level
  self default: (empty,empty,empty)
   
TestType.ax.x (Coordinate):
  top-level default: (1,2,3)
  direct parent default: (1,0,0)
  self default: (empty,empty,empty)
  
TestType.a.x.z (Float)
  top-level default: 3
  direct parent default: empty
  self default: empty


"difvalue" stores all differences with top-level default
"pdif" stores all differences of pdefault with self-default
"""

class model_base(object):
  def __init__(self, typetree,default, toplevel, arrayparent):
    self.typetree = typetree
    self.typename = typetree.typename 
    self.is_file = False
    if self.typename is not None: self.typename +=  "Array" * typetree.arraycount
    if hasattr(typetree, "type") and typetree.type is not None: 
      self.type = typetree.type
    elif self.typename is not None:
      self.type = getattr(Spyder,self.typename)
      if issubclass(self.type, Spyder.File): self.is_file = True
    else:
      self.type = None
    self.is_default = typetree.is_default
    self.default = default
    self.pdefault = None
    if self.is_default and typetree.default_expr is not None: 
      self.pdefault = self.type(typetree.default_expr)    
    if default is None and self.pdefault is not None: self.default = self.pdefault
    self.toplevel = toplevel
    self.arrayparent = arrayparent
    self.maybe_none = (self.is_default == True or self.default is not None or self.arrayparent == True)
    #print("MODEL", self.typename, self.is_default, self.default, self.pdefault, self.maybe_none)

    self.hooks = []
    self.exc = None
    self.exc_hooks = []
    self.listeners = []
  def listen(self, listener):
    self.listeners.append(listener)
  def unlisten(self, listener):
    if listener in self.listeners: 
      self.listeners.remove(listener)    
      return True
    else: return False
  def add_hook(self,hook):
    self.hooks.append(hook)
  def add_exc_hook(self,hook):
    self.exc_hooks.append(hook)
  def _notify(self, arg):
    for hook in self.hooks:
      hook(arg)
    for listener in self.listeners:
      listener(arg)  
  def _exc(self):
    self.exc = sys.exc_info()
    for hook in self.exc_hooks:
      hook(self)
      
class model_basic(model_base):
  def __init__(self, typetree,default=None,toplevel=False,arrayparent=False):
    assert typetree.arraycount == 0
    self.members = typetree.members
    assert self.members is None
    model_base.__init__(self, typetree,default,toplevel,arrayparent)
    self.truevalue = None    
    self.value = self.default
  def set(self, value):
    if self.is_file: 
      if value == "": value = None
      if isinstance(value, tuple) and len(value) == 2:
        if value[0] == None: value = None
        if value[0] == "": value = None        
    if value is None:
      if self.maybe_none:
        if self.truevalue is None: return
        self.value = None
        self.truevalue = None
        self._notify(None)
        return
      else: 
        try:
          raise ValueError(None)
        except ValueError:
          self._exc()
          return

    if self.type is None:
      v = dict(value)
    else:
      try:      
        if self.is_file and \
         (isinstance(value, str) or isinstance(value, tuple)):
          if isinstance(value, str):
            fileclass = Spyder.Data
          elif isinstance(value, tuple):
            value, fileclassname = value
            fileclass = getattr(Spyder, fileclassname)
          v = Spyder.File(value,fileclass,"r")
        else:
          v = self.type(value)
      except Exception:
        self._exc()
        return

    if v == self.value: return
    self.value = v
    vv = v
    if vv == self.default: vv = None
    if vv == self.truevalue: return
    self.truevalue = vv
    self._notify(self.truevalue)      
  def get(self,minimal): return self.value
  def dif(self): return self.truevalue
   
class model_array(model_base):
  def __init__(self, typetree, default=None,toplevel=False,arrayparent=False):
    assert typetree.arraycount > 0
    model_base.__init__(self, typetree, default,toplevel,arrayparent)
    self.difvalue = None
    self.children = []
    self.pdif = []
    self._set_hooks_off = False
    self.value = self.default
    child_typename = self.typename[:-len("Array")]
    self.child_typetree = spyder.core.get_typetreedict(child_typename)
    if self.child_typetree.arraycount > 0: self.child_model = model_array
    elif self.child_typetree.members is None: self.child_model = model_basic
    else: self.child_model = model_complex
    if self.default is not None:
      for mnr,mdefault in enumerate(self.default):
        mod = self.child_model(self.child_typetree, mdefault,arrayparent=True)
        h = partial(self._set_hook, mnr)
        mod.hooks.append(h)
        h = partial(self._exc_hook, mnr)
        mod.exc_hooks.append(h)
        self.children.append(mod)

        pp = None
        if self.pdefault is not None: 
          p = None
          if mnr < len(self.pdefault):
            p = self.pdefault[mnr]            
          if p is not None and mod.pdefault != p:
            pp = p
        self.pdif.append(pp)
        
  def __len__(self): 
    return len(self.children)
  def __getitem__(self, key):
    assert isinstance(key, int)
    assert key >= 0
    assert len(self.children) == len(self.pdif)
    for n in range(len(self.children),key+1):
      mod = self.child_model(self.child_typetree,arrayparent=True)
      h = partial(self._set_hook, n)
      mod.hooks.append(h)
      h = partial(self._exc_hook, n)
      mod.exc_hooks.append(h)      
      self.children.append(mod)

      pp = None
      if self.pdefault is not None: 
        p = None
        if n < len(self.pdefault):
          p = self.pdefault[n]            
        if p is not None and mod.pdefault != p:
          pp = p
      self.pdif.append(pp)
      
    return self.children[key]    
  def _exc_hook(self, membername, member):
    if member.exc is None:
      if self.exc is None: return
      if not isinstance(self.exc, list): return
      if membername > len(self.exc): return
      self.exc[membername] = None
      if self.exc == [None] * len(self.exc):
        self.exc = None        
    else:
      if self.exc is None or not isinstance(self.exc, list): self.exc = []
      for n in range(len(self.exc),membername+1): self.exc.append(None)
      self.exc[membername] = copy_module.copy(member.exc)
    for hook in self.exc_hooks:
      hook(self)

  def _post_set(self):
    d = self.difvalue 
    if d is None and self.maybe_none:
      self._notify(None)
      self.value = None
      return
    
    if d is None and self.value is None: 
      return
   
    d2 = []
    for mnr in range(len(self.children)):
      dd = None
      if self.children[mnr].is_default: dd = self.children[mnr].pdefault
      
      dd2 = self.pdif[mnr]
      if dd2 != None: dd = dd2
      
      if d is not None and mnr < len(d):
        dd2 = d[mnr]
        if dd2 != None: dd = dd2
       
      if dd is None: break
      d2.append(dd)
    
    try:      
      v = self.type(d2)
      self.exc = None      
    except Exception:
      self._exc()
      return
    self.value = v
    
    self._notify(self.value)      
  
  def _set_hook(self, member, mv): 
    if mv is None:
      d = self.difvalue 
      if d is None: return
      for n in range(len(self.difvalue), member+1): self.difvalue.append(None)
      self.difvalue[member] = None
      if self.difvalue == [None] * len(self.difvalue): self.difvalue = None
      
    else:
      if self.difvalue is None:
        self.difvalue = [None] * (member+1)
      else:
        for n in range(len(self.difvalue), member+1): self.difvalue.append(None)
      d = self.difvalue 
      d[member] = mv
    
    #reset global exception, if there was one...
    if isinstance(self.exc, tuple): self.exc = None
    
    if self.exc is not None:
      for n in range(len(self.exc), member+1): self.exc.append(None)      
      self.exc[member] = None
      if self.exc == [None] * len(self.exc): 
        self.exc = None
        for hook in self.exc_hooks:
          hook(self)

    if not self._set_hooks_off: 
      self._post_set()
               
  def set(self, value, copy=False):
    if value is None:
      if self.maybe_none:
        if self.difvalue is None: return
        self.value = None
        self.difvalue = None
        self._notify(None)      
      else:
        try:
          raise ValueError(None)
        except ValueError:
          self._exc()
          return
    try:      
      #v = self.type(value)
      v = self.type()
      for dd in value: v.append(dd)
      v.validate()
      self.exc = None
    except Exception:
      import traceback
      self._exc()
      return

    if v == self.value: return
    self.value = v
    self.exc = []
    self._set_hooks_off = True
    for vvnr,vv in enumerate(v):      
      self[vvnr].set(vv)
    self._set_hooks_off = False
    self._post_set()
    self._notify(self.value)      
  def get(self,minimal): return self.value
  def dif(self): return self.difvalue

class model_complex(model_base):
  def __init__(self, typetree,default=None,toplevel=False,arrayparent=False):
    assert typetree.arraycount == 0
    model_base.__init__(self, typetree,default,toplevel,arrayparent)
    self.members = dict(typetree.members)
    self.membernames = getattr(typetree,"membernames", None)
    if self.membernames is None:
      self.membernames = [m[1] for m in self.type._requiredmembers]
      self.membernames += [m[1] for m in self.type._defaultmembers]
    self.difvalue = None
    self.pdif = {}
    self.children = {}
    self._set_hooks_off = False
    self.value = self.default
    for member in self.members:
      mtypetree = self.members[member]
      if mtypetree.arraycount > 0: mmodel = model_array
      elif mtypetree.members is None: mmodel = model_basic
      else: mmodel = model_complex
      mdefault = None
      if self.default is not None: mdefault = getattr(self.default,member)
      mod = mmodel(mtypetree,mdefault,arrayparent=self.arrayparent)
      h = partial(self._set_hook, member)
      mod.hooks.append(h)
      h = partial(self._exc_hook, member)
      mod.exc_hooks.append(h)
      self.children[member] = mod
      
      p = None
      if self.pdefault is not None: 
        p = getattr(self.pdefault,member, None)
        if p is not None and mod.pdefault != p:
          self.pdif[member] = p
  
  def _exc_hook(self, membername, member):
    if member.exc is None:
      if self.exc is None: return
      if not isinstance(self.exc, dict): return
      if membername not in self.exc: return
      del self.exc[membername]      
      if not len(self.exc):
        self.exc = None        
    else:
      if self.exc is None or not isinstance(self.exc, dict): self.exc = {}
      self.exc[membername] = copy_module.copy(member.exc)
    for hook in self.exc_hooks:
      hook(self)

  def _set_hook(self, member, mv): 
    if mv is None:
      d = self.difvalue 
      if d is None: return
      if member in d:
        if len(d) == 1:
          self.difvalue = None
        else:        
          d.pop(member)
    else:
      if self.difvalue is None:
        self.difvalue = {member:mv}
      else:  
        d = self.difvalue 
        d[member] = mv
    
    if self.exc is not None and member in self.exc:
      del self.exc[member] 
      if not len(self.exc):
        self.exc = None      
      for hook in self.exc_hooks:
        hook(self)

    if not self._set_hooks_off: 
      self._post_set()

  def _post_set(self):  
    d = self.difvalue 
    if d is None and self.maybe_none:
      if not self._set_hooks_off: 
        self._notify(None)
        self.value = None
      return
    if d is None and self.value is None: 
      return

    d2 = {}
    #why?
    #for m in self.membernames: 
    #  if self.children[m].is_default: 
    #    d2[m] = self.children[m].pdefault
    d2.update(self.pdif)
    if d is not None: d2.update(d)

    if self.type is None:
      #we are not a Spyder type, but a dict of parameters:
      # partial dicts are OK to contain!
      self.value = d2     
    else:
      try: 
        ##quickly construct an instance v out of the d2 dict
        v = self.type.empty()
        mod = v.__copydict__(d2)
        mod2 = v.__unpack__(*mod)
        v.__construct__(*mod2)
        for mem,m in zip(self.membernames, mod):
          setattr(v,mem,m)
        v.validate()
        self.exc = None      
      except Exception:
        self._exc()
        return
      if self.value is None:
        self.value = v
      for mem in self.membernames:
        setattr(self.value,mem,getattr(v,mem))      
    
    if self.difvalue is None:
      self._notify(None)
    else:
      self._notify(self.value)
           
  def set(self, value, copy=False):
    if value is None:
      if self.maybe_none:
        if self.difvalue is None: return
        self.value = None
        self.difvalue = None
        for member in self.members:
          self.children[member].set(None)
        self._notify(None)
      else: 
        try:
          raise ValueError(None)
        except ValueError:
          self._exc()
          return
    try:      
      v0 = self.type(value)
    except Exception:
      self._exc()
      return
    
    if copy or self.toplevel:
      v = v0
    else:
      assert type(value) == self.type
      v = value
      
    same = False
    if v == self.value: same = True
    self.value = v
    self.exc = {}
    if same: return
        
    self._set_hooks_off = True

    for member in self.members:
      vv = getattr(v0, member)
      self.children[member].set(vv)

    self._set_hooks_off = False
    self._post_set()
    self._notify(self.value)
  def get(self): 
    return self.value
  def dif(self): return self.difvalue

def model(typ):
  from Spyder import Object
  ok = False
  try:
    if isinstance(typ, spyder.core.typetreeclass): 
      ok = True
      typetree = typ
  except TypeError:
    pass
  try:
    if issubclass(typ, Object):
      ok = True
      typetree = typ._typetree()    
  except TypeError:
    pass
  if ok == False: raise TypeError("argument to model must be either typetree or Spyder class")

  if typetree.arraycount > 0: mo = model_array
  elif typetree.members is None: mo = model_basic
  else: mo = model_complex
  return mo(typetree, toplevel = True)
    
    

