"""
Model-view-controller class

Takes five arguments: spyderform, model, view, controller, comparator

- A spyderform

- A callable that returns a model (formtools.model)
  A model must support .get, .exc and .dif methods at the top level 
   (formtools.model actually provides it at all levels, as well as a .set attribute;
     however, below top level they are all UNRELIABLE and must not be used!)
    .get must return a valid Spyder object, or None
     if .get ever returns a valid Spyder object, it must continue to return one forever
    .exc must return None, a sys.exc_info() tuple, or a (nested) dict 
      with sys.exc_info() tuples at the deepest level (leaves)
    .dif must return None, or a nested dict containing all attributes with values other than their defaults
     (and only those!)

  Finally, the model must support a .listen() method  at the top level
  This method must accept a listener.
  The listener must be called when the model's .get Spyder object changes value,
   with the new value as argument
  Calling .unlisten with a listener argument will remove the listener from the list
   (returning True) or returning False if the listener is not present

- A callable that returns a view to the Spyderform (e.g. a gtkview)
  The view must support the .value attribute and .set, .(un)listen and .(un)block methods 
   on all elementary form elements, i.e. form elements with a .type value other
   than undefined, None, "required", or "none"
   Listeners will be called when the value of the element changes (and only then)  
   (In fact, gtkview supports this at all possible levels, not just the form elements
   This is so that the view can also be changed programmatically, but this is out of the
   scope of the mvc class)
  Calling a .set method must update the widget and activate the listeners if it changed
   
- A callable that returns a controller (formtools.controller)   
  The controller must support a ._set method at each possible level 
  (so that the controller can also be changed programmatically; this is a feature
   of the mvc class)
  
  It is assumed that the controller is already linked to the model,
   in the sense that calling the ._set method will update the model 
    if the manipulation is valid
  The _set method must be able to accept any Spyder-parsable value  (Spyder classes, dicts, etc.)
  Invoking the _set method may never raise any kind of validation error 
   However, if the argument is not Spyder-parsable, the whole ._set invocation will be 
    ignored, only updating the model's .exc attribute
  
  Manipulations of the model via the controller's ._set methods are buffered by the controller 
   until the sum of the manipulations leads to a valid Spyder model

- A callable that returns a comparator
  A comparator must have a ._set attribute at the top level that accepts a valid Spyder object
  It must have a ._get method and a ._listen method method at the level of elementary form elements (and nowhere else)
   The ._listen method must takes a listener as an input.
   It must also have a corresponding _unlisten method
  Invoking ._set will update the values of all underlying elements
   All listeners must be called automatically when the value of the element changes (and only then)   

  After the ._set is invoked, all comparator elements will be updated and, if changed, their registered listeners triggered
  Form elements that are inside an optional element with the value None will not be updated in any way

  It is assumed that the comparator's .get method returns the corresponding element in the model,
   or None if it is not there 
  It is also assumed that the model's .listen event is already connected to the comparator's .set method
  The comparator will then update its values, generating .listen events 
   on the elemental widget level
  

The mvc class has three retrieval methods: .model(), .view() and .controller()

The mvc class has four primary methods: .bind(), .unbind(), .block() and .unblock()

The bind() method tethers the model to the view, to the extent possible under their constraints:
 The model is constrained by Spyder validation, which may never be broken; 
 The model will have an initial value of None; once the value becomes a valid Spyder object,
  it will remain valid forever

 The view will have its values constrained to what its underlying widgets can display 

 Neither view nor model will have its constraints violated by the tethering

 The tethering is achieved as follows:

 - The view is connected to the controller, which will try to update the model
 This is done at the level of the elemental widgets (.listen method)
 Walking the spyderform tree will tell where the elemental widgets are.
  
 - The controller can also be invoked programmatically
 
 - The controller is already connected to the model

 - The model is already connected to the comparator
   The comparator will only generate new events if the value changed

 - The comparator's changes are connected to the view's .set method at the widget level
   This will normally not invoke a new event from the widget, since it already has this value
    unless:
    - The origin of the change was programmatical (direct access to the controller)
    - The validate() function of the model has somehow side effects on the model
  
The .unbind() method will remove the tethering.

The .block() method will temporarily break the tethering
(not suspend it! the tether events will be lost, possibly bringing model and view out of sync!)
The .unblock() method will restore the tethering

Finally, the mvc class has two synchronization methods: sync_v2m and sync_m2v

The .sync_v2m() method will walk the form tree and when an elementary element is encountered,
 the corresponding element in the view is retrieved 
 then, its .value attribute is sent to the ._set method
  of the corresponding element in the controller

The .sync_m2v() method will walk the form tree and when an elementary element is encountered,
 the corresponding the element in the comparator is retrieved 
  it's ._get method is called and its return value is sent to the .set method
   of the corresponding element of the view


"""

from functools import partial

def walk(form, obj1, obj2, visitor, position, ar_fill_none_obj1 = False, ar_fill_none_obj2 = False):
  typ = getattr(form, "type", -1)
  if typ in ("none", None): return  
  if hasattr(form, "_disabled") and form._disabled: return

  if form.arraycount > 0:
    assert hasattr(form,"length") and form.length > 0, position
    for n in range(form.length):
      mform = form[n]
      try:
        mobj1 = obj1[n]
      except IndexError:
        if not ar_fill_none_obj1: raise
        mobj1 = None
      try:
        mobj2 = obj2[n]
      except IndexError:
        if not ar_fill_none_obj2: raise
        mobj2 = None

      mposition = position 
      if len(position): mposition += "-"
      mposition += str(n)
      walk(mform, mobj1, mobj2, visitor, mposition, ar_fill_none_obj1, ar_fill_none_obj2)
  else:    
    membernames = form.get_membernames()
    
    assert obj1 is not None or ar_fill_none_obj1 is True, position
    assert obj2 is not None or ar_fill_none_obj2 is True, position
    leaf = False
    if membernames is None or len(membernames) == 0: leaf = True
    elif typ not in (-1, "none", "required"): leaf = True
    if leaf:
      visitor(form, obj1, obj2, position)
    else:
      for member in membernames:
        mform = getattr(form,member)
        mobj1 = getattr(obj1,member,None)
        mobj2 = getattr(obj2,member,None)
        mposition = position 
        if len(position): mposition += "-"
        mposition += member
        walk(mform, mobj1, mobj2, visitor, mposition, ar_fill_none_obj1, ar_fill_none_obj2)
        

class mvc(object):
  def __init__(self, spyderform, model, view, controller, comparator):
    self.spyderform = spyderform
    self.model = model
    self.view = view
    self.controller = controller
    self.comparator = comparator
    self._vc_listeners = []
    self._cv_listeners = []
    self._blocked = False
  
  def _if_not_blocked(self, callback, *args, **kwargs):
    if self._blocked: return
    callback(*args, **kwargs)  
    
  def _bind_vc_visitor(self, form_leaf, view_leaf, controller_leaf, pos):
    listener = controller_leaf._set    
    listener2 = partial(self._if_not_blocked, listener)
    view_leaf.listen(listener2)
    self._vc_listeners.append((view_leaf.unlisten,listener2))
    self._v2m_listeners.append((lambda: view_leaf.value, listener)) 

  def _bind_cv_visitor(self, form_leaf, comparator_leaf, view_leaf, pos):
    listener = view_leaf.set
    listener2 = partial(self._if_not_blocked, listener)
    comparator_leaf._listen(listener2)
    self._cv_listeners.append((comparator_leaf._unlisten,listener2))  
    self._m2v_listeners.append((comparator_leaf._get, listener)) 
  
  def unbind(self):
    for unlisten, listener in self._vc_listeners:
      unlisten(listener)
    self._vc_listeners = []
    for unlisten, listener in self._cv_listeners:
      unlisten(listener)
    self._cv_listeners = []
    
  def bind(self):
    self.unbind()
    self._v2m_listeners = []
    self._m2v_listeners = []
    walk(self.spyderform, self.view(), self.controller(), self._bind_vc_visitor, "")
    walk(self.spyderform, self.comparator(), self.view(), self._bind_cv_visitor, "")
    
  def block(self): self._blocked = True
  def unblock(self): self._blocked = False
    
  def sync_v2m(self):
    for getter, listener in self._v2m_listeners:
      val = getter()
      listener(val)

  def sync_m2v(self):
    for getter, listener in self._m2v_listeners:
      val = getter()
      listener(val)
  

