"""
Converts a form into a simple_model_controller object
A simple_model_controller object supports indexing and/or attribute access at the lowest level
However, all initial values are set to None

You can change only initial values, and only at the lowest level 
 (forms that have no subforms)
Files are recognized as a special case: their values are set with a string,
  but actually a File object is constructed

Until its value (or the value of one of its members) is set,
 simple_model_controller == None evaluates to True (but simple_model_controller is None remains False!!)

"""

from functools import partial

class simple_model_controller(object):
  def __init__(self, form, has_parent = False):
    self._form = None
    self._value = None
    self._length = 0
    self._struct_properties = None
    self._properties = None
    if form is None: 
      assert has_parent == True
      return
    import Spyder, spyder
    assert isinstance(form, spyder.core.spyderform)
    if hasattr(form, "type") and form.type == "none":
      assert has_parent == True
      return
    self._form = form
    membernames = form.get_membernames()
    self._setter = None
    self._getter = None
    self._is_none = True
    if membernames is None and self._form.arraycount == 0: 
      assert has_parent == True
      assert form.typename is not None
      is_file = False 
      if hasattr(form, "type") and form.type == "file": 
        is_file = True
        self._spyderfiletype = Spyder.File
      if hasattr(form, "spydertype") and form.spydertype is not None:
        spydertype = form.spydertype
      else:
        spydertype = getattr(Spyder, form.typename)
      if issubclass(spydertype, Spyder.File):
        is_file = True
        self._spyderfiletype = spydertype

      self._getter = partial(getattr, self, "_value")
      if is_file:
        self._setter = self.filesetter
      else:
        self._setter = self.vsetter
      return
    if self._form.arraycount > 0:
      assert hasattr(form, "length") and form.length is not None
      membernames = [str(v) for v in range(form.length)]
      getfunc = lambda v: form.__getitem__(int(v))
    else:
      getfunc = partial(getattr, form)
    self._properties = {}
    self._struct_properties = {}
    for membername in membernames:
      f = simple_model_controller(getfunc(membername), True)
      if f._properties is not None: self._struct_properties[membername] = f
      elif f._form is not None: self._properties[membername] = f
      
  def vsetter(self, v):    
    self._is_none = False
    self._value = v
    
  def filesetter(self, filename):
    self._is_none = False
    self._value = self._spyderfiletype(filename, self._form.file, "r") 

  def __getitem__(self, key):    
    if not isinstance(key, int): raise TypeError
    if not key >= 0: raise IndexError
    if not self._form.arraycount > 0: raise TypeError
    try:
      ret = self._struct_properties[str(key)]
    except KeyError:    
      try:
        ret = self._properties[str(key)]
      except KeyError:
        raise IndexError
    if key > self._length: self._length = key
    return ret
  def __setitem__(self, key, value):    
    if not isinstance(key, int): raise TypeError
    if not key >= 0: raise IndexError
    if not self._form.arraycount > 0: raise TypeError
    kkey = str(key)
    assert kkey in self._properties
    self._properties[kkey] = value

  def __getattr__(self, attr):
    try:
      return self._struct_properties[attr]
    except KeyError:
      try:
        return self._properties[attr]._getter()
      except KeyError:
        if self._value is None: 
          raise AttributeError(attr) 
        else:
          return getattr(self._value, attr)
  def __setattr__(self, attr, value):
    if attr.startswith("_"): 
      return object.__setattr__(self, attr, value)
    try:
      ret = self._properties[attr]._setter(value)
      self._is_none = False
      return ret
    except KeyError:
      if self._value is None: 
        raise AttributeError(attr) 
      else:
        return setattr(self._value, attr, value)
  def __eq__(self, other):
    if other is None: return self._is_none
    return object.__eq__(self, other)
  def __ne__(self, other):
    return not self.__eq__(other)
  def __len__(self): return self._length
  def dict(self):
    if self._is_none: return None
    if self._form.arraycount > 0:
      ret = []
      for n in range(self._form.length):
        v = self[n]
        if isinstance(v, simple_model_controller): v = v.dict()
        ret.append(v)
        return ret
    else:
      ret = {}
      for attr in self._struct_properties:
        v = self._struct_properties[attr].dict()
        if v is not None: ret[attr] = v
      for attr in self._properties:  
        v = self._properties[attr]._getter()
        if v is not None: ret[attr] = v
      if not len(ret): return None
      return ret
