#TODO: CGI parser: take care that default values are treated properly! (see old htmlform)

#TODO: in other form generator function, the token generator does not return lines, 
# but returns a dict of lines (GLADE) or receives an additional first argument at the beginning must be modified

#document: count_from_one, [None], memberorder, hr element generator

import spyder
import Spyder
from Spyder import Object, File, String, Integer, Bool, Float
from functools import partial

###############################################################################
# SETTINGS
###############################################################################

form_start = '<form encType="multipart/form-data" action="%s" method="post"><table>\n'
form_submit = """  
  <tr>
    <td colspan="2" align="center" style="width:710px"><input type="submit" /></td>
  </tr>  
"""
form_end = '</table></form>'

tabstep = 2

#if "wrappings" is between these values, the generated HTML depends on "wrappings"
# therefore, if wrappings changes (due to grouping), the HTML must then be regenerated
wrappings_regen_min = 2
wrappings_regen_max = 1000

###############################################################################

_generators = {}
_group_generators = {}
_ele_generators = {}
_phase_generators = {}
_tokens = ("outerwrap", "midwrap", "head", "innerwrap", "group") 
_phases = ("outer", "pre", "post", "mid", "inner")
_allowedformtypes = ("required", "none", "text", "password", "number", "slider", "checkbox", "option", "radio", "file", None)  

def set_generator(token, complexity, callback):
 
  """
  Adds an HTML generator for a data model member, given token and complexity
   token must be one of the standard tokens:
    "outerwrap", "midwrap", "head", "innerwrap", "group"
   generators for other tokens must have their phase specified, see set_phase_generator
  
  Complexity indicates how many nested members a model has. 
  A model consisting purely of basic types has a complexity of 0.
  A model consisting of basic types and of 0-complexity models, has a complexity of 1.
  
  The callback must take the following arguments:
    name: the name of the member (can be None for a top level data model) 
    id: the identifier of the member: AxisSystem.x.y => x-y
    objs:  lists the object or class for which HTML is being generated
      In determining the value of the object, the following lookup priority is used:
        - The .value attribute of the supplied form, if present
        - The supplied object, if not None
        - The .default attribute of the supplied form, if present
        - None
      The objs argument is a stack list: 
      the first item refers to the actual object, the second item to its parent,
       the third one to the parent's parent, etc.
    forms: the data model's spyderform
      This argument is a stack list: 
      the first item refers to the actual form, the second item to its parent,
       the third one to the parent's parent, etc.
      spyderforms are auto-generated by Spyder classes, 
       but custom spyderforms can be supplied as well
    name2: 
      Only supplied for the "head" token
      In determining the value of name2, the following lookup priority is used:
        - The .name attribute of the supplied form, if present
        - The name of the member
    wrappings:
      The number of wrapping elements that the form has
      each parent, group and group element adds a wrapping element
    group: 
      Only supplied for the "group" token: value of the .group attribute
    groupindex: 
      Only supplied for the "group" token: starts at 1, increments for every new group
    mode: 
      Not supplied for the "head" token
      "start" or "end" for the other tokens; also "pre" or "post" for groups and group elements
    elegroup: 
       Only supplied for group elements
       It is possible to do advanced grouping with group elements
       Elegroups are constructed with form.new_group(id, type)
       The form code is then expected to set/append elegroup.members 
        with names of members and IDs of other group elements
    args:
      Not supplied for standard tokens
      For other tokens, the arguments in the function call 
 
  Definition
    Spyderforms are typically defined within the form block of Spyder models
      but other definition is also possible
    For the most part, it is simply the setting of attributes  
    The spyderforms of the members are also contained in the spyderform
    For arrays, you can access the spyderforms of elements with []
    
    Generating data for other tokens is done by add_<token> function calls
    i.e. member.add_header(<text>)
    This requires the registration of a phase generator, mapping "header"
     to a generator AND a certain phase
     
    
  Generation

    First, the htmlform function determines the values of the 
     name, id, objs and forms attributes

    Then, the generators are called in the following order:

      the other tokens of phase "outer", with mode "start"
      "outerwrap" with mode "start"
      the other tokens of phase "pre", with mode "start"
      "head"
      the other tokens of phase "mid", with mode "start"
      "midwrap" with mode "start"
      the other tokens of phase "post", with mode "start"
      "innerwrap" with mode "start"
      the other tokens of phase "inner", with mode "start"
      For every member:
        if it is in a group that has already been processed, skip it
        if it is not in a group:
          select its spyderform and generate HTML 
        else:
          "group" with mode "start"
          For every member in the group:
            The HTML of the member is generated
            If form.type is not in the allowed list, then 
             a type generator must have been defined for that type.
            
          "group" with mode "end"  
      the other tokens of phase "inner", with mode "end"    
      "innerwrap" with mode "end" 
      the other tokens of phase "post", with mode "end"
      "midwrap" with mode "end" 
      the other tokens of phase "mid", with mode "end"
      the other tokens of phase "pre", with mode "end"
      "outerwrap" with mode "start"
      the other tokens of phase "outer", with mode "end"
      
  """
  assert isinstance(complexity, int) and complexity >= 0
  _generators[token, complexity] = callback

def set_group_generator(grouptype, complexity, callback):
  assert isinstance(complexity, int) and complexity >= 0
  _group_generators[grouptype, complexity] = callback

def set_element_generator(element, callback):
  _ele_generators[element] = callback

def set_phase_generator(element, complexity, phase, callback):
  """
  Defines a generator for an extra token
  In addition to the complexity, an appropriate phase must be associated
   with the generator callback
  At generation time, generate_other_tokens is invoked for each phase
  Generate_other_tokens will then call the generator callback, including
   the wrapped arguments and all other parameters, but only if the phase
   corresponds to the phase under which the callback is registered    
  """
  assert phase in _phases, phase
  _phase_generators[element, complexity] = phase, callback
  
from . import generators
generators.init()

from .resolve import resolve

###########################################################################
 
def process_other_tokens(phase, t, html, props, complexity, data):
  if "_othertokens" not in props: return
  name, id, objs, forms, wrappings, mode = data
  tokens = props["_othertokens"]
  if mode == "end": tokens = reversed(tokens)
  
  for token, args in tokens:
    if (token, complexity) not in _phase_generators:
      raise ValueError("Unknown token %s, complexity %s" % (token, complexity))
    tokphase, callback = _phase_generators[token, complexity]
    if tokphase != phase: continue
    ret = callback(name, id, objs, forms, wrappings, mode, args)
    if ret is not None:
      html.append("%s%s" % (t.tabstr, ret))
      if mode == "start": t.incr()
      elif mode == "end": t.decr()
  

def _write_html(objs, forms, name=None,prename="",wrappings=0):  

  obj, form = objs[0], forms[0]

  if hasattr(form, "value"):
    defaultvalue = form.value
  elif obj is not None:
    defaultvalue = obj
  elif hasattr(form, "default"):
    defaultvalue = form.default
  else:
    defaultvalue = None

  class tabclass:
    def __init__(self):
     self.currtabs = 0
     self.tabstr = ""    
    def incr(self):
      self.currtabs += tabstep
      self.tabstr = self.currtabs * " "
    def decr(self):
      self.currtabs -= tabstep
      self.tabstr = self.currtabs * " "
  t = tabclass()
  
  complexity = 0
  mhtmls = []
  mhtmldata = [] #stores some values in case we have to regenerate HTML
  
  n0, n = None, None
  typnam = form.typename + "Array" * form.arraycount
  if name is not None:
    n0 = prename + name
    n = prename.replace(".", "-") + name
 
  
  curr_membernames = None
  if form.arraycount > 0:
    is_array = True
    is_none = False   
    if hasattr(form, "type"):
      if form.type not in ("none", "required"): is_array = False
      if form.type == "none": is_none = True
    if is_array:
      if is_none is False:
        if not hasattr(form, "length"):
          raise TypeError("Cannot generate HTML for \"%s\"(%s): member is an Array, but no length has been specified" % (n0, typnam))      
        if not isinstance(form.length, int) or form.length <= 0:
          raise TypeError("Cannot generate HTML for \"%s\"(%s): member is an Array, but its length is not a positive integer" % (n0, typnam))                  
        curr_membernames = []
        curr_members = []
        curr_objs = []
        for nr in range(form.length):
          curr_membernames.append(str(nr))
          curr_members.append(form[nr])
          mobj = None
          if defaultvalue is not None and len(defaultvalue) > nr:
            mobj = defaultvalue[nr]
          curr_objs.append(mobj)  
  elif form.get_membernames():
    curr_membernames = form.get_membernames()
    curr_members = []
    curr_objs = []
    for membername in curr_membernames:
      curr_members.append(getattr(form, membername))
      mobj = None
      if defaultvalue is not None: 
        mobj = getattr(defaultvalue, membername)
      curr_objs.append(mobj)
      
  mwrappings = wrappings + 1
  if curr_membernames is not None:    
    if hasattr(form, "memberorder"):
      mems = []
      for mnr, m in enumerate(form.memberorder):
        assert m not in form.memberorder[mnr+1:]
        assert m in curr_membernames, m        
        mems.append(curr_membernames.index(m))
      for mnr, membername in enumerate(curr_membernames):        
        if membername not in form.memberorder:
          mems.append(mnr)
      curr_membernames = [curr_membernames[i] for i in mems]      
      curr_members = [curr_members[i] for i in mems]
      curr_objs = [curr_objs[i] for i in mems]
    mcomplexities = []
    for membername, mform, mobj in zip(curr_membernames, curr_members, curr_objs):
      newprename = prename
      if name is not None: 
        newprename += name + "."
      mobjs = [mobj, defaultvalue] + objs[1:]
      mforms = [mform] + forms
      mhtml, mcomplexity = _write_html(mobjs, mforms, membername,newprename,mwrappings)
      mhtmls.append(mhtml)
      mhtmldata.append((mobjs, mforms, newprename))
      if hasattr(mform, "group") and mform.group is not None:
        mcomplexity += 1
      mcomplexities.append(mcomplexity)
    complexity = max(mcomplexities) 
    if form.arraycount == 0 or complexity == 0: complexity +=  1  
 
  for token in _tokens:
    assert token,complexity in _generators
  gen_head = _generators["head", complexity]
  gen_group = _generators["group", complexity]
  gen_innerwrap = _generators["innerwrap", complexity]
  gen_midwrap = _generators["midwrap", complexity]
  gen_outerwrap = _generators["outerwrap", complexity]

  if name is None:
    gen_head = None
    gen_outerwrap = None
    gen_midwrap = None
  html = []
  while 1:
    if name is not None:
      if form.is_default:
        if not hasattr(form, "type") and defaultvalue is None:
          raise TypeError("Cannot generate HTML for \"%s\"(%s): member is optional, but no default value and no type (required, none or other) is specified" % (n0, typnam))
        if hasattr(form, "type") and form.type in ("none", None): break
      else:
        if hasattr(form, "type") and form.type in ("none", None):
          raise TypeError("Cannot generate HTML for \"%s\"(%s): member is not optional, but type is specified as \"none\"" % (n0, typnam))

    tokdata = (name, id, objs, forms, wrappings, "start")

    process_other_tokens("outer", t, html, form._props, complexity, tokdata)

    if gen_outerwrap is not None:
      start = gen_outerwrap(name, n, objs, forms, wrappings, "start")
      if start is not None: html.append(t.tabstr + start)
      t.incr()

    process_other_tokens("pre", t, html, form._props, complexity, tokdata)

    if gen_midwrap is not None:
      start = gen_midwrap(name, n, objs, forms, wrappings, "start")
      if start is not None: html.append(t.tabstr + start)
      t.incr()

    process_other_tokens("mid", t, html, form._props, complexity, tokdata)

    name2 = name
    if hasattr(form, "name"): name2 = form.name
    if gen_head is not None: 
      val = gen_head(name, n, objs, forms, wrappings, name2)
      if val is not None: html.append(t.tabstr + val)

    process_other_tokens("post", t, html, form._props, complexity, tokdata)

    if gen_innerwrap is not None:
      start = gen_innerwrap(name, n, objs, forms, wrappings, "start")
      if start is not None: html.append(t.tabstr + start)
      t.incr()

    process_other_tokens("inner", t, html, form._props, complexity, tokdata)
    
    curr_wrappings = mwrappings
    if complexity > 0:
      done_membernames = set()
      groupindex = 0
      resolve(
       html,
       t,
       curr_wrappings, 
       done_membernames,
       mhtmls,
       mhtmldata,
       mwrappings,
       curr_membernames,
       curr_members,
       mcomplexities,
       groupindex,
       name,
       n,
       objs,
       forms,
       wrappings,
       complexity,
       gen_group,
      )
    else:
      typ = getattr(Spyder, form.typename)
      formtype = "text"
      if issubclass(typ, File):
        if form.arraycount > 0:
          raise TypeError("Cannot generate HTML for \"%s\"(%s): array of files" % (n0, typnam))
        if not hasattr(form, "file"):
          raise TypeError("Cannot generate HTML for \"%s\"(%s): a file type (file attribute) must be specified" % (n0, typnam))
        if hasattr(form, "options"):
          raise TypeError("Cannot generate HTML for \"%s\"(%s): file elements cannot have options" % (n0, typnam))
        formtype = "file"
      elif issubclass(typ, Bool):
        formtype = "checkbox"
       
      if hasattr(form, "options"):
        assert formtype == "text"
        formtype = "option"
        options = form.options      
        optiontitles = options
        if hasattr(form, "optiontitles"):
          optiontitles = form.optiontitles
          assert len(options) == len(optiontitles), (n, len(options), len(optiontitles))

      if hasattr(form, "type"):
        if form.type not in ("required", "none", None):
          formtype = form.type
        
      assert formtype in _allowedformtypes or formtype in _ele_generators, formtype
            
      if formtype in _ele_generators:
        gen = _ele_generators[formtype]
        ret = gen(name, n, objs, forms, wrappings, defaultvalue) 
        html.append(t.tabstr + ret)
      elif formtype == "option":
        assert hasattr(form, "options"), n
        html.append("%s<select name=\"%s\" id=\"%s\">" % (t.tabstr,n, n))
        t.incr()
        if defaultvalue is None:
          html.append("%s<option value=\"\" selected></option>" % (t.tabstr))
        else:
          assert defaultvalue in options, (n, defaultvalue, options)
        for opt, optitle in zip(options, optiontitles):
          sel = ""
          if opt == defaultvalue: sel = " selected"
          html.append("%s<option value=\"%s\"%s>%s</option>" % (t.tabstr, opt, sel, optitle))
        t.decr()
        html.append("%s</select>" % (t.tabstr))
      elif formtype == "radio":
        if defaultvalue is not None:
          assert defaultvalue in options, (n, defaultvalue, options)
        for opt, optitle in zip(options, optiontitles):
          checked = ""
          if opt == defaultvalue: checked = " checked"
          html.append("%s<input type = \"radio\" value=\"%s\"%s>%s" % (t.tabstr, opt, checked, optitle))        
      else:
        defaultstr = ""
        checked = ""
        if formtype == "file":
          """ #does not work in HTML... re-use for some other generator
          if defaultvalue is not None:
            if isinstance(defaultvalue, File):
              defaultstr = defaultvalue.name
            elif isinstance(defaultvalue, str):
              defaultstr = defaultvalue
          """
          pass 
        elif formtype == "checkbox":
          if defaultvalue is not None and defaultvalue == True:
            checked = " checked"
        else:
          assert formtype == "text" or formtype == "password", formtype
          if defaultvalue is not None:
            defaultstr = str(defaultvalue)
        valstr = " value = \"%s\"" % defaultstr
        maxlen = ""
        if hasattr(form, "length"): maxlen = " maxlength=%d" % form.length
        html.append("%s<input type=\"%s\" name=\"%s\" id=\"%s\"%s%s%s>" % (t.tabstr,formtype,n,n,valstr,maxlen,checked))

    tokdata = (name, id, objs, forms, wrappings, "end")
    process_other_tokens("inner", t, html, form._props, complexity, tokdata)
    
    if gen_innerwrap is not None:
      t.decr()
      end = gen_innerwrap(name, n, objs, forms, wrappings, "end")
      if end is not None: html.append(t.tabstr + end)

    process_other_tokens("post", t, html, form._props, complexity, tokdata)
    
    if gen_midwrap is not None:
      t.decr()
      end = gen_midwrap(name, n, objs, forms, wrappings, "end")
      if end is not None: html.append(t.tabstr + end)

    process_other_tokens("mid", t, html, form._props, complexity, tokdata)
    process_other_tokens("pre", t, html, form._props, complexity, tokdata)
    
    if gen_outerwrap is not None:
      t.decr()
      end = gen_outerwrap(name, n, objs, forms, wrappings, "end")
      if end is not None: html.append(t.tabstr + end)

    process_other_tokens("outer", t, html, form._props, complexity, tokdata)
    
    html.append("")
    break
    
  return html, complexity

def htmlform(obj=None, form=None, cgi=None):
  """
  obj can be: A Spyder class, a Spyder object, or None
    if obj is None, then form must not be None
  form is a spyderform object
    if form is None, it is extracted from the Spyder obj
  cgi is the name of the CGI file (string)
  """
  assert cgi is not None
  if form is None:
    if isinstance(obj, Object):
      form = obj._form()
    else:
      ok = False
      try: 
        if issubclass(obj, Object):
          form = obj._form() 
          obj = None
          ok = True
      except TypeError:
        pass
      if not ok: raise TypeError(obj)
  else:
    try:
      if issubclass(obj, Object): 
        obj = None
    except TypeError:
      pass    
  assert isinstance(form, spyder.core.spyderform)
  
  ret = form_start % cgi
  html, complexity = _write_html([obj], [form])
  for l in html:
    ret += " " * tabstep + l + "\n"
  ret += form_submit  
  ret += form_end
  return ret
